/*
 * Copyright (c) 1998 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "common.h"
#pragma hdrstop

#include "chcodes.h"
#include "keymap.h"
#include "keymap.rh"
#include "ktelnet.rh"
#include "emul.rh"
#include "tickets.rh"

#define KTS(key) ((GetKeyState(key) >> 7) & 1)

static int cmds[] = {
CM_SHOWTICKETS, CM_NEWCONN, CM_OPENCONN, CM_CLOSECONN, CM_SETUP, CM_EXIT,
CM_EDITCOPY, CM_EDITPASTE, CM_EDITMARKALL, CM_TELNET_SEND_AYT,
CM_TELNET_SEND_IP, CM_TELNET_SEND_AO, CM_TELNET_SEND_EC, CM_TELNET_SEND_EL,
CM_TELNET_SEND_BRK, CM_KRBLOGIN, CM_KRBCLEAR, CM_KEYMAP_EDIT,
CM_KEYMAP_DEFAULT, CM_KEYMAP_IMPORT, CM_KEYMAP_EXPORT,
CM_SCROLLBASE+SB_LINEUP, CM_SCROLLBASE+SB_LINEDOWN,
CM_SCROLLBASE+SB_PAGEUP, CM_SCROLLBASE+SB_PAGEDOWN,
CM_TERMINAL_RESET, CM_CAPTURE, CM_ENDCAPTURE, CM_SENDTEXT,
CM_PRINTEREJECT, CM_PRINTSELECTED, CM_SERIALSETUP,
-1};

class KeyMapDialog;

class TrackingButton : public TButton
{
      public:
	TrackingButton(KeyMapDialog *parent, int resourceId);

      protected:
	uint EvGetDlgCode(MSG far*);
	void EvKillFocus(HWND);
	void EvKeyDown(uint key, uint repeatCount, uint flags);
	void EvKeyUp(uint key, uint repeatCount, uint flags);
	void EvSysKeyDown(uint key, uint repeatCount, uint flags);
	void EvSysKeyUp(uint key, uint repeatCount, uint flags);

      private:
	KeyMapDialog *kmd;

	DECLARE_RESPONSE_TABLE(TrackingButton);
};

class KeyMapDialog : public TDialog
{
      public:
	KeyMapDialog(TWindow *parent, KeyMap *km);
	~KeyMapDialog();

	void SetupWindow();
	void TrackOff();
	bool Track(uint key, uint repeatCount, uint flags);
	bool GetTracking();

      protected:
	void AddProps(TComboBoxData &cbd, int BaseResource,
		int Min, int Max, bool first = false);
	void AddProps(TComboBoxData &cbd, int *cmds, bool first = false);
	const char *ResString(int ResId);
	bool Changed();
	void SaveCurrent();

	void WinEnable(TEdit *win, bool EnableFlag);
	void WinEnable(TWindow *win, bool EnableFlag);

	KeyMap *KM;

	struct {
		TComboBoxData keys;
		TComboBoxData cmds;
		} trfbuf;

	WXComboBox *keyb;
	TCheckBox *ctrlKey;
	TCheckBox *shiftKey;
	TCheckBox *altKey;
	TRadioButton *opb[3];
	WXComboBox *cmdb;
	TEdit *ssb;
	TStatic *note;

	void CmKSel();
	void CmOk();

	void EvOperChanged();
	void EvKeyChanged();
	void EvCmdChanged();
	LRESULT CtlColorWindow(WPARAM w, LPARAM l);

	void UpdateSelection(bool ClearShiftKeys);

      private:
	char RSbuf[512];	      
	KeyDescriptor *CurrentKey;
	int CurrentKeyCode;
	int CurrentShift;
	bool ChangedFlag;
	bool KeyTracking;
	bool IgnoreNext;

	DECLARE_RESPONSE_TABLE(KeyMapDialog);
};

inline bool
KeyMapDialog::GetTracking()
{
	return KeyTracking;
}

DEFINE_RESPONSE_TABLE1(TrackingButton, TButton)
  EV_WM_GETDLGCODE,
  EV_WM_KEYDOWN,
  EV_WM_KEYUP,
  EV_WM_SYSKEYDOWN,
  EV_WM_SYSKEYUP,
  EV_WM_KILLFOCUS,	  
END_RESPONSE_TABLE;

TrackingButton::TrackingButton(KeyMapDialog *parent, int resourceId)
	: TButton(parent, resourceId)
{
	kmd = parent;
}

void
TrackingButton::EvKillFocus(HWND)
{
	kmd->TrackOff();
}

uint
TrackingButton::EvGetDlgCode(MSG far* msg)
{
	return TButton::EvGetDlgCode(msg) |
		(kmd->GetTracking()? DLGC_WANTALLKEYS: 0);
}

void
TrackingButton::EvKeyUp(uint key, uint repeatCount, uint flags)
{
	if (kmd->Track(key, repeatCount, flags)) {
		Click();
		kmd->TrackOff();
		}
}

void
TrackingButton::EvSysKeyUp(uint key, uint repeatCount, uint flags)
{
	if (kmd->Track(key, repeatCount, flags)) {
		Click();
		kmd->TrackOff();
		}
}

void
TrackingButton::EvSysKeyDown(uint key, uint repeatCount, uint flags)
{
	kmd->Track(key, repeatCount, flags);
}

void
TrackingButton::EvKeyDown(uint key, uint repeatCount, uint flags)
{
	kmd->Track(key, repeatCount, flags);
}

DEFINE_RESPONSE_TABLE1(KeyMapDialog, TDialog)

	EV_BN_CLICKED(IDC_KMCTRL, EvKeyChanged),
	EV_BN_CLICKED(IDC_KMSHIFT, EvKeyChanged),
	EV_BN_CLICKED(IDC_KMALT, EvKeyChanged),
	EV_BN_CLICKED(IDC_KMDWIN, EvOperChanged),
	EV_BN_CLICKED(IDC_KMDCMD, EvOperChanged),
	EV_BN_CLICKED(IDC_KMSS, EvOperChanged),
	EV_MESSAGE(WM_CTLCOLORLISTBOX, CtlColorWindow),
	EV_MESSAGE(WM_CTLCOLOREDIT, CtlColorWindow),
	EV_MESSAGE(WM_CTLCOLORSTATIC, CtlColorWindow),
	EV_COMMAND(IDOK, CmOk),
	EV_COMMAND(IDC_KMKSEL, CmKSel),
	EV_CBN_SELCHANGE(IDC_KMKEY, EvKeyChanged),
	EV_CBN_SELCHANGE(IDC_KMDCOP, EvCmdChanged),
END_RESPONSE_TABLE;


KeyMapDialog::KeyMapDialog(TWindow *parent, KeyMap *km)
	: TDialog(parent, IDD_KEYMAP)
{
	register int i;

	KM = km;
	CurrentKey = 0;
	CurrentKeyCode = -1;
	KeyTracking = false;
	IgnoreNext = true;

	keyb = new WXComboBox(this, IDC_KMKEY);

	ctrlKey = new TCheckBox(this, IDC_KMCTRL);
	ctrlKey->DisableTransfer();
	shiftKey = new TCheckBox(this, IDC_KMSHIFT);
	shiftKey->DisableTransfer();
	altKey = new TCheckBox(this, IDC_KMALT);
	altKey->DisableTransfer();

	opb[0] = new TRadioButton(this, IDC_KMDWIN);
	opb[0]->DisableTransfer();
	opb[1] = new TRadioButton(this, IDC_KMDCMD);
	opb[1]->DisableTransfer();
	cmdb = new WXComboBox(this, IDC_KMDCOP);
	opb[2] = new TRadioButton(this, IDC_KMSS);
	opb[2]->DisableTransfer();
	ssb = new TEdit(this, IDC_KMSSTR);
	ssb->DisableTransfer();

	new TrackingButton(this, IDC_KMKSEL);

	note = new TStatic(this, IDC_KMNOTE);

	AddProps(trfbuf.keys, IDS_KMKEYS, VK_F1, VK_F24, true);
	AddProps(trfbuf.keys, IDS_KMKEYS, 0, VK_F1-1);
	AddProps(trfbuf.keys, IDS_KMKEYS, VK_F24+1, 255);

	for (i = 0; i < KDM_MAXSIZE; i++) {
		if (KM->KeyDesc[i] == NULL)
			continue;
		if (ResString(IDS_KMKEYS + i) != NULL)
			continue;

		trfbuf.keys.AddStringItem(KM->KeyString(this, i), i);
		}

	AddProps(trfbuf.cmds, cmds, true);

	SetTransferBuffer(&trfbuf);
}

KeyMapDialog::~KeyMapDialog()
{
}

void
KeyMapDialog::SetupWindow()
{
	TDialog::SetupWindow();

	UpdateSelection(true);
	note->SetText("");
}

void
KeyMapDialog::TrackOff()
{
	KeyTracking = false;
	note->SetText("");
}

bool
KeyMapDialog::Track(uint key, uint repeatCount, uint flags)
{
	char buf[512];
	const char *cp;
	register int r;

	if (!KeyTracking)
		return false;

	if ((((flags >> 1) ^ flags) & 0x4000) == 0) {
		if (DebugData.dbgKeyMap) {
			sprintf(buf, "%02x %d %s\r\n%08x", key, repeatCount,
				(flags & 0x8000)?"UP":"DOWN", flags);
			note->SetText(buf);
			}
		ctrlKey->SetCheck((GetKeyState(VK_CONTROL) & 0x80)?
			BF_CHECKED: BF_UNCHECKED);

		shiftKey->SetCheck((GetKeyState(VK_SHIFT) & 0x80)?
			BF_CHECKED: BF_UNCHECKED);

		altKey->SetCheck((GetKeyState(VK_MENU) & 0x80)?
			BF_CHECKED: BF_UNCHECKED);

		cp = 0;
		if (!IgnoreNext) {
			if ((key != VK_SHIFT) &&
			    (key != VK_CONTROL) &&
			    (key != VK_MENU) &&
			    (key != VK_CAPITAL) &&
			    (key != VK_NUMLOCK) &&
			    (key != VK_SCROLL)) {

				cp = KM->KeyString(this, key|(flags & 0x100));

				if (keyb->SetSelString(cp, 0) == CB_ERR) {
					r = keyb->AddString(cp);
					keyb->SetItemData(r, key);
					}
				}
			}
		IgnoreNext = false;

		UpdateSelection(false);
		
		if (cp)
			return true;
		}
	return false;
}

void
KeyMapDialog::CmKSel()
{
	KeyTracking = !KeyTracking;

	if (KeyTracking) {
		note->SetText(ResString(IDS_KMTRACK));
		IgnoreNext = true;
		}
	else
		note->SetText("");
}

void
KeyMapDialog::CmOk()
{
	UpdateSelection(false);
	TDialog::CmOk();
}

void
KeyMapDialog::EvCmdChanged()
{
	ChangedFlag = true;
}

void
KeyMapDialog::EvKeyChanged()
{
	TCurrentEvent ev = GetCurrentEvent();

	UpdateSelection((ev.Message == WM_COMMAND) &&
			(HIWORD(ev.Param1) == CBN_SELCHANGE));
}

void
KeyMapDialog::SaveCurrent()
{
	char buf[1024];
	register int i;

	if (CurrentKey == NULL)
		CurrentKey = new KeyDescriptor(&KM->KeyDesc[CurrentKeyCode],
			CurrentShift, KDO_WINDOWS, 0, 0);

	if (opb[0]->GetCheck() == BF_CHECKED)
		CurrentKey->Update(KDO_WINDOWS);
	else if (opb[1]->GetCheck() == BF_CHECKED) {
		if ((i = cmdb->GetSelIndex()) < 0)
			CurrentKey->Update(KDO_WINDOWS);
		else
			CurrentKey->Update(KDO_COMMAND, cmdb->GetItemData(i));
		}
	else {
		buf[ssb->GetText(buf, sizeof(buf)-1)] = 0;
		CurrentKey->SetSendString(buf);
		}
}

void
KeyMapDialog::UpdateSelection(bool ClearShiftKeys)
{
	char buf[1024];
	const char *cp;
	register int i;
	bool SetDefaultAction;

	if (CurrentKeyCode != -1) {
		if (Changed())
			SaveCurrent();
		}

	if (ClearShiftKeys) {
		ctrlKey->SetCheck(BF_UNCHECKED);
		shiftKey->SetCheck(BF_UNCHECKED);
		altKey->SetCheck(BF_UNCHECKED);
		}

	CurrentShift = 0;
	if (ctrlKey->GetCheck() == BF_CHECKED)
		CurrentShift |= KDS_CTRL;
	if (shiftKey->GetCheck() == BF_CHECKED)
		CurrentShift |= KDS_SHIFT;
	if (altKey->GetCheck() == BF_CHECKED)
		CurrentShift |= KDS_ALT;

	if ((i = keyb->GetSelIndex()) < 0) {
		MessageBeep(-1);
		return;
		}
	
	int VirtualKeyCode = keyb->GetItemData(i);

	CurrentKey = NULL;

	if ((VirtualKeyCode >= 0) && (VirtualKeyCode < KDM_MAXSIZE))
		CurrentKeyCode = VirtualKeyCode;
	else
		CurrentKeyCode = -1;

	if (KM->KeyDesc && (CurrentKeyCode != -1)) {
		if (KM->KeyDesc[CurrentKeyCode]) {
			CurrentKey = KM->KeyDesc[CurrentKeyCode]->
				Find(CurrentShift);
			if (ClearShiftKeys && (CurrentKey == NULL)) {
				CurrentKey = KM->KeyDesc[CurrentKeyCode];
				CurrentShift = CurrentKey->GetShiftBits();
				ctrlKey->SetCheck((CurrentShift & KDS_CTRL)?
					BF_CHECKED : BF_UNCHECKED);
				shiftKey->SetCheck((CurrentShift & KDS_SHIFT)?
					BF_CHECKED : BF_UNCHECKED);
				altKey->SetCheck((CurrentShift & KDS_ALT)?
					BF_CHECKED : BF_UNCHECKED);
				}
			}
		}

	SetDefaultAction = true;
	if (CurrentKey != NULL) {
		switch (CurrentKey->GetOperation()) {
			case KDO_COMMAND:
				SetDefaultAction = false;
				opb[0]->SetCheck(BF_UNCHECKED);
				opb[1]->SetCheck(BF_CHECKED);
				opb[2]->SetCheck(BF_UNCHECKED);
				if ((cp = ResString(
					CurrentKey->GetOparg1())) != NULL) {
					cmdb->SetSelString(cp, 0);
					}
				else
					cmdb->SetSelIndex(-1);
				break;

			case KDO_SEND:
				SetDefaultAction = false;
				opb[0]->SetCheck(BF_UNCHECKED);
				opb[1]->SetCheck(BF_UNCHECKED);
				opb[2]->SetCheck(BF_CHECKED);
				ssb->SetText(
					CurrentKey->GetSendString(buf,
								sizeof(buf)));

				break;
			}
		}

	if (SetDefaultAction) {
		opb[0]->SetCheck(BF_CHECKED);
		opb[1]->SetCheck(BF_UNCHECKED);
		opb[2]->SetCheck(BF_UNCHECKED);
		cmdb->SetSelIndex(-1);
		ssb->SetText("");
		}

	ssb->ClearModify();
	EvOperChanged();
	ChangedFlag = false;
}

bool
KeyMapDialog::Changed()
{
	if (ChangedFlag)
		return true;
	
	if (opb[2]->GetCheck() == BF_CHECKED) {
		if (ssb->IsModified())
			return true;
		}

	return false;
}

void
KeyMapDialog::EvOperChanged()
{
	ChangedFlag = true;

	if (opb[0]->GetCheck() == BF_CHECKED) {
		WinEnable(ssb, false);
		WinEnable(cmdb, false);
		}
	else if (opb[1]->GetCheck() == BF_CHECKED) {
		WinEnable(cmdb, true);
		WinEnable(ssb, false);
		}
	else {
		WinEnable(cmdb, false);
		WinEnable(ssb, true);
		}

}

void
KeyMapDialog::WinEnable(TWindow *win, bool EnableFlag)
{
	if (EnableFlag)
		win->SetStyle(win->GetStyle() & ~WS_DISABLED);
	else
		win->SetStyle(win->GetStyle() | WS_DISABLED);
	win->Invalidate(false);
}

void
KeyMapDialog::WinEnable(TEdit *win, bool EnableFlag)
{
	ssb->SetReadOnly(!EnableFlag);
	WinEnable((TWindow *)win, EnableFlag);
}

LRESULT
KeyMapDialog::CtlColorWindow(WPARAM w, LPARAM l)
{
	LOGBRUSH lb;

	if ((::GetWindowLong((HWND)l, GWL_STYLE) & WS_DISABLED) == 0)
		return DefWindowProc(GetCurrentEvent().Message, w, l);

	lb.lbStyle = BS_SOLID;
	lb.lbColor = GetSysColor(COLOR_3DFACE);
	lb.lbHatch = 0;

	SetBkColor((HDC)w, GetSysColor(COLOR_3DFACE));
	SetTextColor((HDC)w, GetSysColor(COLOR_GRAYTEXT));

	return (LRESULT)::CreateBrushIndirect(&lb);
}

void
KeyMapDialog::AddProps(TComboBoxData &cbd, int BaseResource,
int Min, int Max, bool first)
{
	register int i;
	register const char *cp;
	
	for (i = Min; i <= Max; i++) {
		if ((cp = ResString(i + BaseResource)) != NULL) {
			cbd.AddStringItem(cp, i, first);
			first = false;
			}
		if ((cp = ResString(i + BaseResource + 0x100)) != NULL) {
			cbd.AddStringItem(cp, i + 0x100, first);
			first = false;
			}
		}
}

void
KeyMapDialog::AddProps(TComboBoxData &cbd, int *cmds, bool first)
{
	register const char *cp;
	
	for (;*cmds != -1; cmds++) {
		if ((cp = ResString(*cmds)) != NULL) {
			cbd.AddStringItem(cp, *cmds, first);
			first = false;
			}
		}
}

const char *
KeyMapDialog::ResString(int ResId)
{
	if (GetApplication()->LoadString(ResId, RSbuf, sizeof(RSbuf)-1) == 0)
		return NULL;

	return RSbuf;
}

//****************************************************************

KeyMap::KeyMap()
{
	KeyDesc = 0;
}

KeyMap::~KeyMap()
{
	Clear();

	if (KeyDesc)
		delete[] KeyDesc;
}

void
KeyMap::Clear()
{
	register int i;

	if (KeyDesc) {
		for (i = 0; i < KDM_MAXSIZE; i++) {
			if (KeyDesc[i])
				delete KeyDesc[i];
			KeyDesc[i] = 0;
			}
		}
}
int
KeyMap::Edit(TWindow *parent)
{
	return KeyMapDialog(parent, this).Execute();
}

void
KeyMap::Export(TWindow *parent)
{
	FILE *fp;
	register int i;
	KeyDescriptor *kd;
	char buf[128];
	const char *cp;
	int res;
	KTFileData FileData;

	FileData.Flags |= OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;
	FileData.SetFilter(ResString(parent, IDS_KMEXTS));
	FileData.DefExt = "MAP";

	if (TFileSaveDialog(parent, FileData, 0,
		ResString(parent, IDS_KMEXPTO)).Execute() != IDOK)
		return;

	if ((fp = fopen(FileData.FileName, "w")) == NULL) {
		FileError(parent, FileData.FileName, IDS_EXPORTERR);
		return;
		}

	HCURSOR OldCur = ::SetCursor(::LoadCursor(0, IDC_WAIT));

	for (i = 0; i < KDM_MAXSIZE; i++) {
		if (KeyDesc[i] == NULL)
			continue;
		fprintf(fp, "# %s\n", KeyString(parent, i));
		for (kd = KeyDesc[i]; kd; kd = kd->GetNext()) {
			fprintf(fp, "0x%03x ", i);
			fprintf(fp, "%-5s",
				(kd->GetShiftBits() & KDS_CTRL)?"Ctrl":"");
			fprintf(fp, "%-6s",
				(kd->GetShiftBits() & KDS_SHIFT)?"Shift":"");
			fprintf(fp, "%-3s\t",
				(kd->GetShiftBits() & KDS_ALT)?"Alt":"");
			
			switch (kd->GetOperation()) {
				case KDO_WINDOWS:
					fprintf(fp, "WINDOWS\n");
					break;
				case KDO_COMMAND:
					cp = ResString(parent, kd->GetOparg1());
					if (*cp == 0) {
						sprintf(buf, "%d", kd->GetOparg1());
						cp = buf;
						}

					fprintf(fp, "COMMAND\t%s\n", cp);
					break;
				case KDO_SEND:
					fprintf(fp, "SEND\t%s\n",
						kd->GetSendString(buf,
							sizeof(buf), true));
					break;
				case KDO_DEAD:
					fprintf(fp, "DEAD\n");
					break;
				default:
					fprintf(fp, "???? %d\n",
						kd->GetOperation());
				}
			}
		}
	
	res = ferror(fp);
	res |= fclose(fp);
	
	::SetCursor(OldCur);

	if (res)
		FileError(parent, FileData.FileName, IDS_EXPORTERR);
}

bool
KeyMap::Import(TWindow *parent)
{
	FILE *fp;
	register int i, cod, linenum, OpArg1;
	KeyDescriptor *kd;
	char buf[512], *pp, *tok;
	KTFileData FileData;
	char *whitespace = " \t";
	int shift;
	enum _kdoper OpCode;

	FileData.SetFilter(ResString(parent, IDS_KMEXTS));
	FileData.Flags |= OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;
	FileData.DefExt = "MAP";

	if (TFileOpenDialog(parent, FileData, 0,
		ResString(parent, IDS_KMIMPFROM)).Execute() != IDOK)
		return false;

	if ((fp = fopen(FileData.FileName, "r")) == NULL) {
		FileError(parent, FileData.FileName, IDS_IMPORTERR);
		return false;
		}

	switch (parent->MessageBox(ResString(parent, IDS_KMCLR), header,
		MB_YESNOCANCEL|MB_ICONQUESTION)) {
		case IDYES:
			Clear();
			break;

		case IDCANCEL:
			fclose(fp);
			return false;
		}

	HCURSOR OldCur = ::SetCursor(::LoadCursor(0, IDC_WAIT));

	for (linenum = 1; !feof(fp) ; linenum++) {
		if (fgets(buf, sizeof(buf), fp) == NULL)
			break;

		i = strlen(buf)-1;
		if (buf[i] != '\n') {
			i = IDS_KMILTL;
			goto syntax;
			}
		else
			buf[i] = 0;

		pp = buf;
		while (isspace(*pp))
			pp++;

		if ((*pp == '#') || (*pp == ';'))
			continue;		// Comment.
				
		cod = strtol(pp, &pp, 0);
		shift = 0;
		
		if ((cod <= 0) || (cod >= KDM_MAXSIZE)) {
			i = IDS_KMICODE;
			goto syntax;
			}

		while ((tok = strtok_r(NULL, whitespace, &pp)) != NULL) {
			if (strcasecmp(tok, "Ctrl") == 0)
				shift |= KDS_CTRL;
			else if (strcasecmp(tok, "Shift") == 0)
				shift |= KDS_SHIFT;
			else if (strcasecmp(tok, "Alt") == 0)
				shift |= KDS_ALT;
			else if (strcasecmp(tok, "WINDOWS") == 0) {
				OpCode = KDO_WINDOWS;
				break;
				}
			else if (strcasecmp(tok, "COMMAND") == 0) {
				OpCode = KDO_COMMAND;
				break;
				}
			else if (strcasecmp(tok, "SEND") == 0) {
				OpCode = KDO_SEND;
				break;
				}
			else if (strcasecmp(tok, "DEAD") == 0) {
				OpCode = KDO_DEAD;
				break;
				}
			else {
				i = IDS_IMPSYNTAX;
				goto syntax;
				}
			}
		while (isspace(*pp))
			pp++;

		if (OpCode == KDO_COMMAND) {
			for (i = 0; cmds[i] != -1; i++) {
				if (strcasecmp(pp,
					ResString(parent, cmds[i])) == 0)
					break;
				}
			if (cmds[i] == -1) {
				i = IDS_KMICMDE;
				goto syntax;
				}
			
			OpArg1 = cmds[i];
			}
		
		if (KeyDesc[cod])
			kd = KeyDesc[cod]->Find(shift);
		else
			kd = 0;
		
		if (!kd)
			kd = new KeyDescriptor(&KeyDesc[cod], shift, KDO_WINDOWS, 0, 0);
		
		if (OpCode == KDO_SEND)
			kd->SetSendString(pp);
		else
			kd->Update(OpCode, OpArg1, 0);
		}

	if (ferror(fp)) {
		FileError(parent, FileData.FileName, IDS_IMPORTERR);
		fclose(fp);
		return false;
		}

	fclose(fp);
	return true;
	
      syntax:
	sprintf(buf, ResString(parent, i), linenum, FileData.FileName);
	::SetCursor(OldCur);
	parent->MessageBox(buf, errorHeader, MB_OK|MB_ICONEXCLAMATION);
	fclose(fp);
	return false;
}

void
KeyMap::FileError(TWindow *parent, const char *fn, int resId)
{
	char buf[1024];

	sprintf(buf, ResString(parent, resId), fn, Wstrerror(errno));

	parent->MessageBox(buf, errorHeader, MB_OK|MB_ICONEXCLAMATION);
}

bool
KeyMap::LoadFromRegistry(Registry *reg)
{
	register int i;
	int size;
	char *data = 0;
	int CharCode;
	int ShiftBits;
	enum _kdoper OpCode;
	int Arg1;
	char *Arg2;
	
	try {
		reg->Open();
		if ((size = reg->GetBinSize("KeyMap")) != 0) {
			data = new char[size];
			reg->GetBinData("KeyMap", (BYTE *)data, size);
			}
		reg->Close();
		}
	catch (RegistryFail) {
		reg->Close();
		if (data)
			delete[] data;
		size = 0;
		}

	if (size == 0)
		return false;
	
	for (i = 0; i < size;) {
		if (i >= (size - sizeof(short)))
			break;
		CharCode = *(short *)&data[i];
		i += sizeof(short);
		if (i >= size)
			break;
		ShiftBits = data[i++];
		if (i >= size)
			break;
		OpCode = (enum _kdoper)data[i++];
		if (i > (size - sizeof(int)))
			break;
		Arg1 = *(int *)&data[i];
		i += sizeof(int);
		if (OpCode == KDO_SEND) {
			Arg2 = &data[i];
			i += Arg1;
			}
		else
			Arg2 = 0;
		if (i > size)
			break;
		
		if ((OpCode == KDO_COMMAND) && (Arg1 == CM_KEYMAP_EDIT_OLD))
			Arg1 = CM_KEYMAP_EDIT;	// See note in emul.rh.

		_AddSingleKeyMap(CharCode, ShiftBits, OpCode, Arg1, Arg2);
		}

	delete[] data;
	return true;
}

void
KeyMap::SaveInRegistry(Registry *reg)
{
	register int i;
	int size;
	char *data;

	size = 0;
	for (i = 0; i < KDM_MAXSIZE; i++) {
		if (KeyDesc[i] == NULL)
			continue;
		size += KeyDesc[i]->BinSave(i, NULL);
		}

	data = new char[size];

	size = 0;
	for (i = 0; i < KDM_MAXSIZE; i++) {
		if (KeyDesc[i] == NULL)
			continue;
		size += KeyDesc[i]->BinSave(i, &data[size]);
		}

	try {
		reg->Open(true);
		reg->SetBinData("KeyMap", (BYTE *)data, size);
		reg->Close();
		}
	catch (RegistryFail) {
		reg->Close();
		}

	delete[] data;
}

void
KeyMap::_AddSingleKeyMap(int VirtualCode,
int ShiftBits,
enum _kdoper OpCode, int OpArg1, const char *OpArg2)
{
	if ((VirtualCode < 0) || (VirtualCode >= KDM_MAXSIZE))
		return;

	if (KeyDesc == 0) {
		KeyDesc = new KeyDescriptor *[KDM_MAXSIZE];
		ZeroMemory(KeyDesc, sizeof(KeyDescriptor *) * KDM_MAXSIZE);
		}

	new KeyDescriptor(&KeyDesc[VirtualCode],
		ShiftBits, OpCode, OpArg1, OpArg2);
}


bool
KeyMap::Process(int key, TWindow *t)
{
	if ((key >= 0) && (key < KDM_MAXSIZE)) {
		if (KeyDesc[key]) {
			int Shift = 0;
			if (GetKeyState(VK_CONTROL) & 0x80)
				Shift |= KDS_CTRL;
			if (GetKeyState(VK_SHIFT) & 0x80)
				Shift |= KDS_SHIFT;
			if (GetKeyState(VK_MENU) & 0x80)
				Shift |= KDS_ALT;
			
			if (KeyDesc[key]->Process(Shift, t))
				return true;
			}
		}
	return false;
}

const char *
KeyMap::KeyString(TWindow *parent, int key)
{
	register int r;
	const char *cp;
	
	cp = ResString(parent, IDS_KMKEYS + key);
	
	if (*cp)
		return cp;		// Simple.

	cp = ResString(parent, IDS_KMKEYS + (key & 0xff));
	if (*cp)
		return cp;		// Almost simple.

	if ((key >= 'A') && (key <= 'Z'))
		r = IDS_KMLETTER;
	else if ((key >= '0') && (key <= '9'))
		r = IDS_KMDIGIT;
	else
		r = IDS_KMCODE;
	
	sprintf(KSbuf, ResString(parent, r), key);
	return KSbuf;
}

//****************************************************************

KeyDescriptor::KeyDescriptor(KeyDescriptor **Link, 
		      int ShiftBits,
		      enum _kdoper OpCode,
		      int OpArg1,
	              const char *OpArg2)
{
	shiftBits = ShiftBits;
	opCode = OpCode;
	opArg1 = OpArg1;

	if ((OpCode == KDO_SEND) && opArg1 && OpArg2) {
		opArg2 = new char[opArg1];
		CopyMemory((void *)opArg2, OpArg2, opArg1);
		}
	else
		opArg2 = 0;

	next = *Link;
	*Link = this;
}

KeyDescriptor::~KeyDescriptor()
{
	if (opArg2)
		delete[] opArg2;

	if (next)
		delete next;
}

KeyDescriptor *
KeyDescriptor::Find(int Shift)
{
	if (Shift == shiftBits)
		return this;
	
	if (next)
		return next->Find(Shift);
	
	return NULL;
}

const char *
KeyDescriptor::GetSendString(char *buf, int bufSize, bool FirstSpace)
{
	register int i, j;
	
	if (opCode != KDO_SEND) {
		*buf = 0;
		return buf;
		}

	for (i = j = 0; (j < (bufSize-5)) && (i < opArg1); i++) {
		switch (opArg2[i]) {
			case BEL:
				buf[j++] = '\\';
				buf[j++] = 'a';
				break;
			case BS:
				buf[j++] = '\\';
				buf[j++] = 'b';
				break;
			case HT:
				buf[j++] = '\\';
				buf[j++] = 't';
				break;
			case Linefeed:
				buf[j++] = '\\';
				buf[j++] = 'n';
				break;
			case Return:
				buf[j++] = '\\';
				buf[j++] = 'r';
				break;
			case ESC:
				buf[j++] = '\\';
				buf[j++] = 'e';
				break;

			default:
				if (((opArg2[i] & 0xff) < ' ') ||
				    ((opArg2[i] & 0x7f) == 0x7f) ||
				    ((opArg2[i] == ' ') &&
				     (i == 0) &&
				     (FirstSpace))) {
					int v = opArg2[i] & 0xff;
					buf[j++] = '\\';
					buf[j++] = (char)
						(((v >> 6) & 7) + '0');
					buf[j++] = (char)
						(((v >> 3) & 7) + '0');
					buf[j++] = (char)
						(((v >> 0) & 7) + '0');
					}
				else
					buf[j++] = opArg2[i];
			}
		}
	buf[j] = 0;
	
	return buf;
}

void
KeyDescriptor::SetSendString(char *buf)
{
	register int i, j, v;

	for (i = j = 0; buf[i];) {
		if (buf[i] == '\\') {	// Special!
			i++;
			switch (buf[i]) {
				case 'a':
					buf[j++] = BEL;
					i++;
					continue;
				case 'b':
					buf[j++] = BS;
					i++;
					continue;
				case 't':
					buf[j++] = HT;
					i++;
					continue;
				case 'n':
					buf[j++] = Linefeed;
					i++;
					continue;
				case 'r':
					buf[j++] = Return;
					i++;
					continue;
				case 'e':
					buf[j++] = ESC;
					i++;
					continue;

				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
					v = buf[i++] - '0';
					if ((buf[i] >= '0') &&
					    (buf[i] <= '7'))
						v = (v << 3) + buf[i++] - '0';
					if ((buf[i] >= '0') &&
					    (buf[i] <= '7'))
						v = (v << 3) + buf[i] - '0';
					buf[j++] = (char)v;
					if ((buf[i] >= '0') &&
					    (buf[i] <= '7')) {
						i++;
						continue;
						}
					break;
				}
			}

		if (buf[i])
			buf[j++] = buf[i++];
		}

	Update(KDO_SEND, j, buf);
}

int
KeyDescriptor::BinSave(int CharCode, char *buf)
{
	int res;
	
	switch (opCode) {
		case KDO_COMMAND:
			res = sizeof(short) + sizeof(char) + sizeof(char) +
				sizeof(int);
			if (buf) {
				*(short *)buf = (short)CharCode;
				buf += sizeof(short);
				*buf++ = (char)shiftBits;
				*buf++ = KDO_COMMAND;
				*(int *)buf = opArg1;
				buf += sizeof(int);
				}
			break;
		case KDO_SEND:
			res = sizeof(short) + sizeof(char) + sizeof(char) + 
				sizeof(int) + opArg1;
			if (buf) {
				*(short *)buf = (short)CharCode;
				buf += sizeof(short);
				*buf++ = (char)shiftBits;
				*buf++ = KDO_SEND;
				*(int *)buf = opArg1;
				buf += sizeof(int);
				CopyMemory((void *)buf, opArg2, opArg1);
				buf += opArg1;
				}
			break;
		default:
			res = 0;
			break;
		}
	
	if (next)
		return next->BinSave(CharCode, buf) + res;
	
	return res;
}

void
KeyDescriptor::Update(enum _kdoper OpCode, int OpArg1, const char *OpArg2)
{
	if (opArg2) {
		delete[] opArg2;
		opArg2 = 0;
		}
	
	opCode = OpCode;
	opArg1 = OpArg1;

	if ((OpCode == KDO_SEND) && opArg1 && OpArg2) {
		opArg2 = new char[opArg1];
		CopyMemory((void *)opArg2, OpArg2, opArg1);
		}
	else
		opArg2 = 0;
}

bool
KeyDescriptor::Process(int Shift, TWindow *t)
{
	if (Shift == shiftBits) {
		switch (opCode) {
			case KDO_COMMAND:
				t->PostMessage(WM_COMMAND, opArg1, 0);
				return true;
			case KDO_SEND:
				if (opArg1 && opArg2)
					t->PostMessage(
						::RegisterWindowMessage(
							WM_REG_SENDSTRING),
						opArg1,
						(LPARAM)opArg2);
				return true;
			case KDO_DEAD:			// Dead Key
				return true;
			}
		return false;
		}
	
	if (next)
		return next->Process(Shift, t);
	
	return false;
}

